"""Tools for working with nested structures of lists, tuples or dictionaries."""

from collections.abc import Generator, Sequence
from types import EllipsisType
from typing import Any, Optional

from jaxtyping import ArrayLike, PyTree


def walk_tree(tree: PyTree) -> Generator[tuple[list, Any]]:
    """Walk through a tree and yield the indices and values of the leaves.

    Parameters
    ----------
    tree :
        The tree to walk through. Can only be a nested tree of lists, tuples and/or
        dictionaries.

    Yields
    ------
    indices, value
        The indices as a list and values of the leaves.
    """
    cursor = []
    generators_list = []
    value = tree
    key = None
    while True:
        if isinstance(value, dict | list | tuple):
            cursor.append(key)
            if isinstance(value, dict):
                generator = iter(value.items())
            else:
                generator = enumerate(value)
            generators_list.append(generator)

        key, value = next(generators_list[-1], (None, None))
        cursor[-1] = key
        if key is None:
            generators_list.pop()
            cursor.pop()
            if len(generators_list) == 0:
                break
            continue
        elif not isinstance(value, dict | list | tuple):
            yield cursor, value


def nested_indexing(
    tree: PyTree,
    indices: str | int | Sequence[str | int],
    add: Optional[ArrayLike] = None,
    at: int | EllipsisType | slice | None | Sequence[EllipsisType | slice | int] = None,
) -> Any:
    """Return the element of a nested structure of lists or tuples.

    Parameters
    ----------
    tree :
        The nested structure of lists or tuples.
    indices :
        The indices of the element to return.
    add : optional
        The element to add to the nested structure of lists or tuples.
    at : optional
        Specifies the position where to add the element.

    Returns
    -------
    element
        The element of the nested structure of lists or tuples.

    """
    element = tree
    if not isinstance(indices, tuple | list):
        indices = [indices]
    for depth, index in enumerate(indices):
        if depth == len(indices) - 1:
            if add is not None:
                if at is not None:
                    element[index] = element[index].at[at].add(add)
                else:
                    element[index] = element[index] + add
        element = element[index]
    return element



#!/usr/bin/env python
#
# icomo documentation build configuration file, created by
# sphinx-quickstart on Fri Jun  9 13:47:02 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another
# directory, add these directories to sys.path here. If the directory is
# relative to the documentation root, use os.path.abspath to make it
# absolute, like shown here.
#

# -- General configuration ---------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = "1.0"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named "sphinx.ext.*") or your custom ones.

from importlib.metadata import version

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx.ext.mathjax",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.viewcode",
    #    "m2r",
    "myst_parser",
    "nbsphinx",
    "sphinx.ext.intersphinx",
]


# Mock imports
autodoc_mock_imports = [
    "numpy",
    "matplotlib",
    "matplotlib.pyplot",
    "mpl_toolkits",
    "scipy",
    "pymc",
    "pytensor",
    "pandas",
    "filelock",
    "arviz",
    "optax",
    "jax",
    "jax.numpy",
    "diffrax",
    "pytensor",
    "equinox",
    "jaxtyping",
    "jax.typing",
]

nbsphinx_allow_errors = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = [".rst", ".md"]

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "ICoMo Toolbox"
copyright = "2023, Jonas Dehning"
author = "Jonas Dehning"

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#


release = version("icomo")
# for example take major/minor
version = ".".join(release.split(".")[:2])


# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", ".ipynb_checkpoints"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output -------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_book_theme"

html_show_sourcelink = False


# Theme options are theme-specific and customize the look and feel of a
# theme further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

html_theme_options = {
    "collapse_navigation": False,
    "navigation_depth": 4,
    "home_page_in_toc": True,
    "show_navbar_depth": 3,
    "show_toc_level": 3,
    "repository_url": "https://github.com/Priesemann-Group/icomo",
    "use_source_button": False,
    "use_edit_page_button": True,
    "use_repository_button": False,
    "use_issues_button": True,
    "use_download_button": True,
    "path_to_docs": "docs",
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/Priesemann-Group/icomo",  # required
            # Icon class (if "type": "fontawesome"), or path to local image
            # (if "type": "local")
            "icon": "fa-brands fa-github",
            "type": "fontawesome",
        }
    ],
    "header_links_before_dropdown": 2,
}


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

# Napoleon settings
# napoleon_google_docstring = True
# napoleon_numpy_docstring = True
napoleon_include_init_with_doc = True
# napoleon_include_private_with_doc = False
# napoleon_include_special_with_doc = True
# napoleon_use_admonition_for_examples = False
# napoleon_use_admonition_for_notes = False
# napoleon_use_admonition_for_references = False
napoleon_use_ivar = True  # True enables links for attribute types
napoleon_use_param = True  # True enables links for parameter types
napoleon_use_rtype = False


# html_logo = "./images/logo.png"
# html_theme_options = {
#    'logo_only': False,
# }
html_title = f"ICoMo Toolbox\n{release}"


# -- Options for HTMLHelp output ---------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "icomodoc"


# -- Options for LaTeX output ------------------------------------------

latex_elements = {
    # The paper size ("letterpaper" or "a4paper").
    #
    # "papersize": "letterpaper",
    # The font size ("10pt", "11pt" or "12pt").
    #
    # "pointsize": "10pt",
    # Additional stuff for the LaTeX preamble.
    #
    # "preamble": "",
    # Latex figure (float) alignment
    #
    # "figure_align": "htbp",
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "icomo.tex",
        "Inference of Compartmental Models Toolbox Documentation",
        "Jonas Dehning",
        "manual",
    ),
]


# -- Options for manual page output ------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        master_doc,
        "icomo",
        "Inference of Compartmental Models Toolbox Documentation",
        [author],
        1,
    )
]


# -- Options for Texinfo output ----------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "icomo",
        "Inference of Compartmental Models Toolbox Documentation",
        author,
        "icomo",
        "One line description of project.",
        "Miscellaneous",
    ),
]


# -- Options for intersphinx -------------------------------------------

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://docs.scipy.org/doc/numpy", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "pymc": ("https://www.pymc.io/projects/docs/en/stable", None),
    "pytensor": ("https://pytensor.readthedocs.io/en/latest/", None),
    "pandas": ("https://pandas.pydata.org/docs/", None),
    "arviz": ("https://python.arviz.org/en/stable/", None),
    "jax": ("https://jax.readthedocs.io/en/latest/", None),
    "diffrax": ("https://docs.kidger.site/diffrax/", None),
    "optax": ("https://optax.readthedocs.io/en/latest/", None),
    "jaxtyping": ("https://docs.kidger.site/jaxtyping/", None),
    "equinox": ("https://docs.kidger.site/equinox/", None),
}

imgmath_image_format = "svg"
imgmath_font_size = 14

autodoc_member_order = "bysource"
autoclass_content = "both"
typehints_use_signature = False
typehints_use_signature_return_annotation = False
typehints_defaults = "comma"
typehints_fully_qualified = False

nitpicky = False

always_use_bars_union = True
autodoc_typehints = "description"


always_document_param_types = True

html_favicon = "images/icomo_icon.png"
html_logo = "images/icomo_logo250px.png"



"""Auxiliary functions."""

import pymc as pm
import pytensor.tensor as pt


def hierarchical_priors(
    name,
    dims,
    beta=1,
    fix_hyper_sigma=None,
    dist_values=pm.Normal,
    centered_parametrization=False,
    empirical_sigma=False,
):
    """Create hierarchical priors for a variable.

    Create an n-dimensional variable with hierarchical prior with name `name` and
    dimensions `dims`. The prior is a normal distribution with standard deviation
    `sigma` which is sample from a half-cauchy distribution with beta parameter `beta`.

    Parameters
    ----------
    name : str
        Name under which the variable is stored in pm.Model
    dims : tuple of str
        Dimensions over which the variable is defined. Define a dimension by passing
        coords={dim_name: np.arange(size)} to pm.Model at creation.
    beta : float, default=1
        Beta parameter of the half-cauchy distribution
    fix_hyper_sigma : float, default=None
        If not None, the standard deviation from which the variable is sampled is fixed
    dist_values : :class:`pymc.Distribution`, default=pm.Normal
        Distribution from which the values are sampled. Can for example be
        functools.partial(pm.StudentT, nu=4) to sample from a StudentT distribution for
        a more robust model.
    centered_parametrization : bool, default=False
        Whether to use a centered or non-centered parametrization for the hierarchical
        prior
    empirical_sigma : bool, default=False
        Whether to set the standard deviation of the normal distribution to the standard
        deviation of the sampled value. Corresponds to an empirical Bayes approach.

    Returns
    -------
    values : pm.Variable
        pm.Variable for the variable with name `name`

    """
    if empirical_sigma:
        if not dist_values == pm.Normal:
            raise RuntimeError("Empirical sigma only works with normal distribution")
    if not empirical_sigma:
        sigma = (
            pm.HalfCauchy(f"sigma_{name}", beta=beta)
            if fix_hyper_sigma is None
            else fix_hyper_sigma
        )
    if not centered_parametrization:
        values = (dist_values(f"{name}_raw", mu=0, sigma=1, dims=dims)) * sigma
        values = pm.Deterministic(f"{name}", values, dims=dims)

    else:
        values = dist_values(f"{name}", mu=0, sigma=sigma, dims=dims)
    if empirical_sigma:
        values = pm.Flat(f"{name}", dims=dims)
        sigma = pt.std(values)
        logp_reg = pm.logp(dist_values.dist(mu=values, sigma=sigma), 0)
        pm.Potential(f"{name}_regularization", logp_reg, dims=dims)

    return values



"""Wrapper of diffrax functions for convenience."""

import inspect
from collections.abc import Callable
from typing import Optional

import diffrax
import jax
from jaxtyping import Array, ArrayLike, PyTree


def diffeqsolve(
    *args,
    ts_out: Optional[ArrayLike] = None,
    ODE: Callable[[ArrayLike, PyTree, PyTree], PyTree] | None = None,
    fixed_step_size: bool = False,
    **kwargs,
) -> diffrax.Solution:
    """Solves a system of differential equations.

    Wrapper of :meth:`diffrax.diffeqsolve`.
    It accepts the same parameters as :meth:`diffrax.diffeqsolve`.
    Additionally, for convenience, it allows the specification of the output
    timesteps ``ts_out`` and
    the system of differential equations ``ODE`` as keyword arguments. If not specified,
    it uses the :class:`diffrax.Tsit5` solver. For a simple ODE, it is enough to
    specify the following keyword arguments:

    Parameters
    ----------
    ts_out :
        The timesteps at which the output is returned. Equivalent to
        ``diffrax.diffeqsolve(..., saveat=diffrax.SaveAt(ts=ts_out))``. It sets
        additionally the initial time ``t0``, the final time ``t1`` and the time step
        ``dt0`` of the solver if not specified separately.
    ODE :
        The function `f(t, y, args)` that returns the derivatives of the variables of
        the system of differential equations. Equivalent to
        ``diffrax.diffeqsolve(..., terms=diffrax.ODETerm(ODE))``.
    y0 :
        The initial values of the variables of the system of differential equations.
    args : PyTree of ArrayLike or Callable
        The arguments of the system of differential equations. Passed as the third
        argument to the ODE function.

    Other Parameters
    ----------------
    fixed_step_size :
        If True, the solver uses a fixed step size of ``dt0``, i.e it uses
        ``stepsize_controller=diffrax.ConstantStepSize()``.

    Returns
    -------
    sol :
        The solution of the system of differential equations. sol.ys contains the
        variables of the system of differential equations at the output timesteps.

    """
    signature = inspect.signature(diffrax.diffeqsolve)
    signature_bound = signature.bind_partial(*args, **kwargs)

    if fixed_step_size:
        if "stepsize_controller" in signature_bound.arguments:
            raise TypeError(
                "Don't simultaneously specify `icomo.diffeqsolve(..., "
                "fixed_step_size=True, "
                "stepsize_controller=...)`"
            )
        kwargs["stepsize_controller"] = diffrax.ConstantStepSize()

    if ts_out is not None:
        if "saveat" in signature_bound.arguments:
            raise TypeError(
                "Don't simultaneously specify `icomo.diffeqsolve(..., "
                "ts_out=..., saveat=...)`"
            )
        kwargs["saveat"] = diffrax.SaveAt(ts=ts_out)
        if "t0" not in signature_bound.arguments:
            kwargs["t0"] = ts_out[0]
        if "t1" not in signature_bound.arguments:
            kwargs["t1"] = ts_out[-1]
        if "dt0" not in signature_bound.arguments:
            kwargs["dt0"] = ts_out[1] - ts_out[0]
    else:
        if "t0" not in signature_bound.arguments:
            raise TypeError(
                "Specify `icomo.diffeqsolve(..., t0=...)` and/or "
                "`icomo.diffeqsolve(..., ts_out=...)`"
            )
        if "t1" not in signature_bound.arguments:
            raise TypeError(
                "Specify `icomo.diffeqsolve(..., t1=...)` and/or "
                "`icomo.diffeqsolve(..., ts_out=...)`"
            )
        if "dt0" not in signature_bound.arguments:
            raise TypeError(
                "Specify `icomo.diffeqsolve(..., dt0=...)` and/or "
                "`icomo.diffeqsolve(..., ts_out=...)`, or "
                "`icomo.diffeqsolve(..., ts_solver=...)`"
            )

    if ODE is not None:
        if "terms" in signature_bound.arguments:
            raise TypeError(
                "Don't simultaneously specify `icomo.diffeqsolve(..., "
                "ODE=..., terms=...)`"
            )
        kwargs["terms"] = diffrax.ODETerm(ODE)
    else:
        if "terms" not in signature_bound.arguments:
            raise TypeError(
                "Specify either `icomo.diffeqsolve(..., ODE=...)` or "
                "`icomo.diffeqsolve(..., terms=...)`"
            )

    if "solver" not in signature_bound.arguments:
        kwargs["solver"] = diffrax.Tsit5()

    return diffrax.diffeqsolve(*args, **kwargs)


def interpolate_func(
    ts_in: ArrayLike,
    values: ArrayLike,
    method: str = "cubic",
    ret_gradients: bool = False,
) -> Callable[[ArrayLike], Array]:
    """
    Return a diffrax-interpolation function that can be used to interpolate pytensors.

    Wrapper of :class:`diffrax.CubicInterpolation` and
    :class:`diffrax.LinearInterpolation`.

    Parameters
    ----------
    ts_in :
        The timesteps at which the time-dependent variable is given.
    vales :
        The time-dependent variable.
    method :
        The interpolation method used. Can be "cubic" or "linear".
    ret_gradients :
        If True, the function returns the gradient of the interpolation function.

    Returns
    -------
    interp :
        The interpolation function. Call ``interp(t)`` to evaluate the
        interpolated variable at time `t`. `t` can be a float or an ArrayLike.

    """
    ts_in = jax.numpy.array(ts_in)
    if method == "cubic":
        coeffs = diffrax.backward_hermite_coefficients(ts_in, values)
        interp = diffrax.CubicInterpolation(ts_in, coeffs)
    elif method == "linear":
        interp = diffrax.LinearInterpolation(ts_in, values)
    else:
        raise RuntimeError(
            f'Interpolation method {method} not known, possibilities are "cubic" or '
            f'"linear"'
        )
    if ret_gradients:
        # return jax.vmap(interp.derivative, 0, 0)
        return interp.derivative
    else:
        # return jax.vmap(interp.evaluate, 0, 0)
        return interp.evaluate



"""Helper functions to model a slow modulation of a time series."""

import jax.numpy as jnp
import numpy as np
import pymc as pm
import pytensor.tensor as pt
from pytensor.link.jax.dispatch.basic import jax_funcify
from pytensor.tensor.sort import ArgSortOp

from .tools import hierarchical_priors


def sigmoidal_changepoints(
    ts_out, positions_cp, magnitudes_cp, durations_cp, reorder_cps=False
):
    r"""Modulation of a time series by sigmoidal changepoints.

    The changepoints are defined by their position, magnitude and duration. The
    resulting equation is:

    .. math::
     f(t) = \sum_{i=1}^{\mathrm{num_cps}} \frac{\mathrm{magnitudes\_cp}[i]}{1 +
     exp(-4 \cdot \mathrm{slope}[i] \cdot (t - \mathrm{positions_cp}[i]))}
    ..

    where slope[i] = magnitudes_cp[i] / durations_cp[i].

    Parameters
    ----------
    t_out: 1d-array
        timepoints where modulation is evaluated, shape: `(time, )`
    t_cp: nd-array
        timepoints of the changepoints, shape: `(num_cps, further dims...)`
    magnitudes: nd-array
        magnitude of the changepoints, shape: `(num_cps, further dims...)`
    durations: nd-array
        magnitude of the changepoints, shape: `(num_cps, further dims...)`
    reorder_cps: bool, default=False
        reorder changepoints such that their timepoints are linearly increasing

    Returns
    -------
    modulation_t: (n+1)d-array
        shape: `(time, further dims...)`

    """
    if reorder_cps:
        order = pt.argsort(positions_cp, axis=0)
        positions_cp = positions_cp[order, ...]
        magnitudes_cp = magnitudes_cp[order, ...]
        durations_cp = durations_cp[order, ...]

    # add necessary empty dimensions to time axis
    ts_out = np.expand_dims(ts_out, axis=tuple(range(1, max(1, positions_cp.ndim) + 1)))
    slope_cp = 1 / durations_cp
    modulation_t = (
        pt.sigmoid((ts_out - positions_cp) * slope_cp * 4) * magnitudes_cp
    )  # 4*slope_cp because the derivative of the sigmoid at zero is 1/4, we want to set
    # it to slope_cp

    modulation_t = pt.sum(modulation_t, axis=1)

    return modulation_t


def priors_for_cps(
    cp_dim,
    time_dim,
    name_positions,
    name_magnitudes,
    name_durations,
    beta_magnitude=1,
    sigma_magnitude_fix=None,
    dist_magnitudes=pm.Normal,
    absolute_magnitude_parametrization=False,
    empirical_bayes_hyper_sigma=False,
    centered_parametrization=False,
    model=None,
):
    """Create priors for changepoints.

    Their positions are uniformly distributed between
    the first and last timepoint. The magnitudes are sampled from a hierarchical prior
    with a beta distribution. The durations are sampled from a normal distribution with
    mean equal to the mean distance between changepoints and standard deviation equal to
    the standard deviation of the distance between changepoints.

    Parameters
    ----------
    cp_dim : str
        Dimension of the :class:`pymc.Model` for the changepoints. Define it by passing
        `coords={cp_dim: np.arange(num_cps)}` to :class:`pymc.Model` at creation. The
        length of this dimension determines the number of changepoints.
    time_dim : str
        Dimension of the :class:`pymc.Model` for the time.
    name_positions : str
        Name under which the positions of the changepoints are stored in
        :class:`pymc.Model`
    name_magnitudes : str
        Name under which the magnitudes of the changepoints are stored in
        :class:`pymc.Model`
    name_durations : str
        Name under which the durations of the changepoints are stored in
        :class:`pymc.Model`
    beta_magnitude : float, default=1
        Beta parameter of the hierarchical prior for the magnitudes
    sigma_magnitude_fix : float, default=None
        If not `None`, the standard deviation from which the magnitudes are sampled is
        fixed
    dist_magnitudes : :class:`pymc.Distribution`, default=pm.Normal
        Distribution from which the magnitudes are sampled. Can for example be
        functools.partial(pm.StudentT, nu=4) to sample from a StudentT distribution for
        a more robust model.
    absolute_magnitude_parametrization : bool, default=False
        Whether to use an parametrization that is absolute or relative to previous
        values for the magnitudes.
    empirical_bayes_hyper_sigma : bool, default=False
        Whether to set the standard deviation of the hierarchical magnitudes to the
        maximum likelihood estimate instead of sampling. Corresponds to an empirical
        Bayes approach.
    centered_parametrization : bool, default=False
        Whether to use a centered parametrization for the hierarchical priors of the
        magnitudes.
    model : :class:`pymc.Model`, default=None
        pm.Model in which the priors are created. If None, the pm.Model is taken from
        the context.

    Returns
    -------
    positions, magnitudes, durations : :class:`pytensor.Variable`
        Variables of dim `cp_dim` that define the positions, magnitudes and durations of
        the changepoints

    """
    model = pm.modelcontext(model)

    time_arr = model.coords[time_dim]
    num_cps = len(model.coords[cp_dim])
    interval_cps = (max(time_arr) - min(time_arr)) / (num_cps + 1)

    ### Positions
    std_Delta_pos = interval_cps / 3
    Delta_pos = pm.Normal(f"Delta_{name_positions}", 0, std_Delta_pos, dims=(cp_dim,))
    positions = Delta_pos + np.arange(1, num_cps + 1) * interval_cps
    pm.Deterministic(f"{name_positions}", positions)

    ### Magnitudes:
    magnitudes_tmp = hierarchical_priors(
        name_magnitudes,
        dims=(cp_dim,),
        beta=beta_magnitude,
        fix_hyper_sigma=sigma_magnitude_fix,
        dist_values=dist_magnitudes,
        centered_parametrization=centered_parametrization,
        empirical_sigma=empirical_bayes_hyper_sigma,
    )
    if not absolute_magnitude_parametrization:
        magnitudes = magnitudes_tmp
    else:
        magnitudes = pt.diff(pt.concatenate([[0.0], magnitudes_tmp]))

    ### Durations:
    mean_duration_len = interval_cps / 3
    std_duration_len = interval_cps / 6
    softplus_scaling = interval_cps / 12
    durations_raw = pm.Normal(
        f"{name_durations}_raw", mean_duration_len, std_duration_len, dims=(cp_dim,)
    )
    durations = pm.Deterministic(
        f"{name_durations}",
        pt.softplus(durations_raw / softplus_scaling) * softplus_scaling,
    )

    order = pt.argsort(positions, axis=0)
    positions = positions[order, ...]
    magnitudes = magnitudes[order, ...]
    durations = durations[order, ...]

    return positions, magnitudes, durations


@jax_funcify.register(ArgSortOp)
def _jax_funcify_Argsort(op, node, **kwargs):
    def argsort(x, axis=None):
        return jnp.argsort(x, axis=axis)

    return argsort



# mypy: disable-error-code="attr-defined"
# isort: skip_file
"""Inference of Compartmental Models Toolbox."""

from importlib import metadata as importlib_metadata

from . import comp_model
from .comp_model import CompModel, erlang_kernel, delayed_copy_kernel

from .diffrax_wrapper import diffeqsolve, interpolate_func

from .jax2pytensor import jax2pytensor

from .tree_tools import nested_indexing, walk_tree

from . import experimental


def _get_version():
    try:
        return importlib_metadata.version("icomo")
    except importlib_metadata.PackageNotFoundError:
        return "unknown"


__author__ = "Jonas Dehning"
__email__ = "jonas.dehning@ds.mpg.de"
__version__ = _get_version()



"""Convert a jax function to a pytensor compatible function."""

import functools as ft
import logging
from collections.abc import Sequence

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
import pytensor.compile.builders
import pytensor.tensor as pt
from jax.tree_util import tree_flatten, tree_map, tree_unflatten
from pytensor.gradient import DisconnectedType
from pytensor.graph import Apply, Op
from pytensor.link.jax.dispatch import jax_funcify

log = logging.getLogger(__name__)

_filter_ptvars = lambda x: isinstance(x, pt.Variable)


def jax2pytensor(jaxfunc, name=None):
    """Return a Pytensor from a JAX jittable function.

    This decorator transforms any JAX jittable function into a function that accepts
    and returns `pytensor.Variables`. The jax jittable function can accept any
    nested python structure (pytrees) as input, and return any nested Python structure.

    It requires to define the output types of the returned values as pytensor types. A
    unique name should also be passed in case the name of the jaxfunc is identical to
    some other node. The design of this function is based on
    https://www.pymc-labs.io/blog-posts/jax-functions-in-pymc-3-quick-examples/

    Parameters
    ----------
    jaxfunc : jax jittable function
        function for which the node is created, can return multiple tensors as a tuple.
        It is required that all return values are able to transformed to
        pytensor.Variable.
    name: str
        Name of the created pytensor Op, defaults to the name of the passed function.
        Only used internally in the pytensor graph.

    Returns
    -------
        A function which can be used in a pymc.Model as function, is differentiable
        and the resulting model can be compiled either with the default C backend, or
        the JAX backend.

    Examples
    --------
    A simple example on how the :func:`jax.numpy.sum` function is used in a PyMC model.

    >>> import jax.numpy as jnp                                # doctest: +ELLIPSIS
    >>> import pymc as pm                                      # doctest: +ELLIPSIS
    >>> import icomo                                           # doctest: +ELLIPSIS
    >>> import numpyro                                         # doctest: +ELLIPSIS
    >>> numpyro.set_host_device_count(4)
    >>> with pm.Model() as model:
    ...     x = pm.Normal("input", mu=1, size=3)
    ...     sum_pt = icomo.jax2pytensor(jnp.sum)
    ...     sum_x = sum_pt(x)
    ...     obs = pm.Normal("obs", sum_x, observed=3)
    >>> trace = pm.sample(model=model, nuts_sampler="numpyro") # doctest: +ELLIPSIS
    >>> print(np.round(np.mean(trace.posterior["input"].to_numpy()),1))
    1.0

    Or a more complex example. One might want to build a model which includes the
    solution of non-linear equation. For instance the intersection of `log(x)` and
    `1/x`:

    >>> import optimistix
    >>> with pm.Model() as model:
    ...     arg1 = pm.HalfNormal("arg1")
    ...     arg1 = pm.math.clip(arg1,0.1,10)
    ...     f1 = lambda x, arg1: jnp.log(x*arg1)
    ...     f2 = lambda x: 1/x
    ...     @icomo.jax2pytensor
    ...     def find_intersection(funcs, arg1):
    ...         f1, f2 = funcs["f1"], funcs["f2"]
    ...         loss = lambda x, _: (f1(x, arg1) - f2(x))**2
    ...               # Loss is squared to make it more convex
    ...         res = optimistix.minimise(fn = loss,
    ...                                   solver=optimistix.BFGS(rtol=1e-8, atol=1e-8),
    ...                                   y0=3.)
    ...         return res
    ...
    ...     intersection_res = find_intersection({"f1": f1, "f2": f2}, arg1)
    ...     intersection = pm.Deterministic("inters", intersection_res.value)
    ...     obs = pm.Normal("obs", mu=intersection, sigma=0.1, observed=3)
    >>> trace = pm.sample(model = model, nuts_sampler="numpyro") # doctest: +ELLIPSIS
    >>> print(f"Inters. = {np.round(np.mean(trace.posterior['inters'].to_numpy()),1)}")
    Inters. = 3.0
    >>> print(f"Std. int. = {np.round(np.std(trace.posterior['inters'].to_numpy()),1)}")
    Std. int. = 0.1

    Jax2pytensor also wraps returned functions from transformed functions, such that
    they can be used either in another transformed function, or evaluated directly to
    obtain the pytensor result:

    >>> with pm.Model() as model:
    ...     arg1 = pm.HalfNormal("arg1")
    ...     arg1 = pm.math.clip(arg1,0.1,10)
    ...     @icomo.jax2pytensor
    ...     def f1_creator(arg1):
    ...         def f_log(x):
    ...             return jnp.log(x*arg1)
    ...         return f_log
    ...     f1 = f1_creator(arg1)
    ...
    ...     f2 = lambda x: 1/x
    ...
    ...     @icomo.jax2pytensor
    ...     def find_intersection(f1, f2):
    ...         loss = lambda x, _: (f1(x) - f2(x))**2
    ...         res = optimistix.minimise(fn = loss,
    ...                                   solver=optimistix.BFGS(rtol=1e-8, atol=1e-8),
    ...                                   y0=3.)
    ...         return res
    ...
    ...     # Pass wrapped f2 function to find_intersection, it creates a pytensor op
    ...     # that also includes arg1 as its input.
    ...     intersection_res = find_intersection(f1, f2)
    ...     intersection = pm.Deterministic("inters", intersection_res.value)
    ...
    ...     # Or evaluate f2 directly
    ...     log_var = pm.Deterministic("log_var", f2(intersection))
    ...     obs = pm.Normal("obs", mu=intersection, sigma=0.1, observed=3)
    >>> trace = pm.sample(model = model, nuts_sampler="numpyro")   # doctest: +ELLIPSIS
    >>> print(f"Inters. = {np.round(np.mean(trace.posterior['inters'].to_numpy()),1)}")
    Inters. = 3.0
    >>> print(f"Std. int. = {np.round(np.std(trace.posterior['inters'].to_numpy()),1)}")
    Std. int. = 0.1
    >>> print(f"log_var = {np.round(np.mean(trace.posterior['log_var'].to_numpy()),1)}")
    log_var = 0.3
    >>> # It also works with the default sampler backend
    >>> trc2 = pm.sample(model = model, cores=1, progressbar=False) # doctest: +ELLIPSIS
    >>> print(f"log_var = {np.round(np.mean(trc2.posterior['log_var'].to_numpy()),1)}")
    log_var = 0.3

    This feature of wrapping returned functions hasn't been tested extensively,
    so please report any issues you might encounter.

    Notes
    -----
    The function is based on a blog post by Ricardo Vieira and Adrian Seyboldt,
    available at
    `pymc-labls.io <https://www.pymc-labs.io/blog-posts/jax-functions-in-pymc-3-quick
    -examples/>`__.
    To accept functions and non pytensor variables as input, the function make use
    of :func:`equinox.partition` and :func:`equinox.combine` to split and combine the
    variables. Shapes are inferred using
    :func:`pytensor.compile.builders.infer_shape` and :func:`jax.eval_shape`.
    """

    def func(*args, **kwargs):
        """Return a pytensor from a jax jittable function."""
        ### Split variables: in the ones that will be transformed to JAX inputs,
        ### pytensor.Variables; _WrappedFunc, that are functions that have been returned
        ### from a transformed function; and the rest, static variables that are not
        ### transformed.

        pt_vars, static_vars_tmp = eqx.partition(
            (args, kwargs), _filter_ptvars, is_leaf=callable
        )
        # is_leaf=callable is used, as libraries like diffrax or equinox might return
        # functions that are still seen as a nested pytree structure. We consider them
        # as wrappable functions, that will be wrapped with _WrappedFunc.

        func_vars, static_vars = eqx.partition(
            static_vars_tmp, lambda x: isinstance(x, _WrappedFunc), is_leaf=callable
        )
        vars_from_func = tree_map(lambda x: x.get_vars(), func_vars)
        pt_vars = dict(vars=pt_vars, vars_from_func=vars_from_func)
        """
        def func_unwrapped(vars_all, static_vars):
            vars, vars_from_func = vars_all["vars"], vars_all["vars_from_func"]
            func_vars_evaled = tree_map(
                lambda x, y: x.get_func_with_vars(y), func_vars, vars_from_func
            )
            args, kwargs = eqx.combine(vars, static_vars, func_vars_evaled)
            return self.jaxfunc(*args, **kwargs)
        """

        pt_vars_flat, vars_treedef = tree_flatten(pt_vars)
        pt_vars_types_flat = [var.type for var in pt_vars_flat]
        shapes_vars_flat = pytensor.compile.builders.infer_shape(pt_vars_flat, (), ())
        shapes_vars = tree_unflatten(vars_treedef, shapes_vars_flat)

        dummy_inputs_jax = jax.tree_util.tree_map(
            lambda var, shape: jnp.empty(
                [int(dim.eval()) for dim in shape], dtype=var.type.dtype
            ),
            pt_vars,
            shapes_vars,
        )

        # Combine the static variables with the inputs, and split them again in the
        # output. Static variables don't take part in the graph, or might be a
        # a function that is returned.
        jaxfunc_partitioned, static_out_dic = _partition_jaxfunc(
            jaxfunc, static_vars, func_vars
        )

        func_flattened = _flatten_func(jaxfunc_partitioned, vars_treedef)

        jaxtypes_outvars = jax.eval_shape(
            ft.partial(jaxfunc_partitioned, vars=dummy_inputs_jax),
        )

        jaxtypes_outvars_flat, outvars_treedef = tree_flatten(jaxtypes_outvars)

        pttypes_outvars = [
            pt.TensorType(dtype=var.dtype, shape=var.shape)
            for var in jaxtypes_outvars_flat
        ]

        ### Call the function that accepts flat inputs, which in turn calls the one that
        ### combines the inputs and static variables.
        jitted_sol_op_jax = jax.jit(func_flattened)
        len_gz = len(pttypes_outvars)

        vjp_sol_op_jax = _get_vjp_sol_op_jax(func_flattened, len_gz)
        jitted_vjp_sol_op_jax = jax.jit(vjp_sol_op_jax)

        if name is None:
            curr_name = jaxfunc.__name__
        else:
            curr_name = name

        # Get classes that creates a Pytensor Op out of our function that accept
        # flattened inputs. They are created each time, to set a custom name for the
        # class.
        SolOp, VJPSolOp = _return_pytensor_ops_classes(curr_name)

        local_op = SolOp(
            vars_treedef,
            outvars_treedef,
            input_types=pt_vars_types_flat,
            output_types=pttypes_outvars,
            jitted_sol_op_jax=jitted_sol_op_jax,
            jitted_vjp_sol_op_jax=jitted_vjp_sol_op_jax,
        )

        @jax_funcify.register(SolOp)
        def sol_op_jax_funcify(op, **kwargs):
            return local_op.perform_jax

        @jax_funcify.register(VJPSolOp)
        def vjp_sol_op_jax_funcify(op, **kwargs):
            return local_op.vjp_sol_op.perform_jax

        ### Evaluate the Pytensor Op and return unflattened results
        output_flat = local_op(*pt_vars_flat)
        if not isinstance(output_flat, Sequence):
            output_flat = [output_flat]  # tree_unflatten expects a sequence.
        outvars = tree_unflatten(outvars_treedef, output_flat)

        static_outfuncs, static_outvars = eqx.partition(
            static_out_dic["out"], callable, is_leaf=callable
        )

        static_outfuncs_flat, treedef_outfuncs = jax.tree_util.tree_flatten(
            static_outfuncs, is_leaf=callable
        )
        for i_func, _ in enumerate(static_outfuncs_flat):
            static_outfuncs_flat[i_func] = _WrappedFunc(
                jaxfunc, i_func, *args, **kwargs
            )

        static_outfuncs = jax.tree_util.tree_unflatten(
            treedef_outfuncs, static_outfuncs_flat
        )
        static_vars = eqx.combine(static_outfuncs, static_outvars, is_leaf=callable)

        output = eqx.combine(outvars, static_vars, is_leaf=callable)

        return output

    return func


class _WrappedFunc:
    def __init__(self, exterior_func, i_func, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.i_func = i_func
        vars, static_vars = eqx.partition(
            (self.args, self.kwargs), _filter_ptvars, is_leaf=callable
        )
        self.vars = vars
        self.static_vars = static_vars
        self.exterior_func = exterior_func

    def __call__(self, *args, **kwargs):
        # If called, assume that args and kwargs are pytensors, so return the result
        # as pytensors.
        def f(func, *args, **kwargs):
            res = func(*args, **kwargs)
            return res

        return jax2pytensor(f)(self, *args, **kwargs)

    def get_vars(self):
        return self.vars

    def get_func_with_vars(self, vars):
        # Use other variables than the saved ones, to generate the function. This
        # is used to transform vars externally from pytensor to JAX, and use the
        # then create the function which is returned.

        args, kwargs = eqx.combine(vars, self.static_vars, is_leaf=callable)
        output = self.exterior_func(*args, **kwargs)
        outfuncs, _ = eqx.partition(output, callable, is_leaf=callable)
        outfuncs_flat, _ = jax.tree_util.tree_flatten(outfuncs, is_leaf=callable)
        interior_func = outfuncs_flat[self.i_func]
        return interior_func


def _get_vjp_sol_op_jax(jaxfunc, len_gz):
    def vjp_sol_op_jax(args):
        y0 = args[:-len_gz]
        gz = args[-len_gz:]
        if len(gz) == 1:
            gz = gz[0]
        func = lambda *inputs: jaxfunc(inputs)
        primals, vjp_fn = jax.vjp(func, *y0)
        gz = tree_map(
            lambda g, primal: jnp.broadcast_to(g, jnp.shape(primal)),
            gz,
            primals,
        )
        if len(y0) == 1:
            return vjp_fn(gz)[0]
        else:
            return tuple(vjp_fn(gz))

    return vjp_sol_op_jax


def _partition_jaxfunc(jaxfunc, static_vars, func_vars):
    """Partition the jax function into static and non-static variables.

    Returns a function that accepts only non-static variables and returns the non-static
    variables. The returned static variables are stored in a dictionary and returned,
    to allow the referencing after creating the function

    Additionally wrapped functions saved in func_vars are regenerated with
    vars["vars_from_func"] as input, to allow the transformation of the variables.
    """
    static_out_dic = {"out": None}

    def jaxfunc_partitioned(vars):
        vars, vars_from_func = vars["vars"], vars["vars_from_func"]
        func_vars_evaled = tree_map(
            lambda x, y: x.get_func_with_vars(y), func_vars, vars_from_func
        )
        args, kwargs = eqx.combine(
            vars, static_vars, func_vars_evaled, is_leaf=callable
        )

        out = jaxfunc(*args, **kwargs)
        outvars, static_out = eqx.partition(out, eqx.is_array, is_leaf=callable)
        static_out_dic["out"] = static_out
        return outvars

    return jaxfunc_partitioned, static_out_dic


### Construct the function that accepts flat inputs and returns flat outputs.
def _flatten_func(jaxfunc, vars_treedef):
    def func_flattened(vars_flat):
        vars = tree_unflatten(vars_treedef, vars_flat)
        outvars = jaxfunc(vars)
        outvars_flat, _ = tree_flatten(outvars)
        return _normalize_flat_output(outvars_flat)

    return func_flattened


def _normalize_flat_output(output):
    if len(output) > 1:
        return tuple(
            output
        )  # Transform to tuple because jax makes a difference between
        # tuple and list and not pytensor
    else:
        return output[0]


def _return_pytensor_ops_classes(name):
    class SolOp(Op):
        def __init__(
            self,
            input_treedef,
            output_treeedef,
            input_types,
            output_types,
            jitted_sol_op_jax,
            jitted_vjp_sol_op_jax,
        ):
            self.vjp_sol_op = None
            self.input_treedef = input_treedef
            self.output_treedef = output_treeedef
            self.input_types = input_types
            self.output_types = output_types
            self.jitted_sol_op_jax = jitted_sol_op_jax
            self.jitted_vjp_sol_op_jax = jitted_vjp_sol_op_jax

        def make_node(self, *inputs):
            self.num_inputs = len(inputs)

            # Define our output variables
            outputs = [pt.as_tensor_variable(type()) for type in self.output_types]
            self.num_outputs = len(outputs)

            self.vjp_sol_op = VJPSolOp(
                self.input_treedef,
                self.input_types,
                self.jitted_vjp_sol_op_jax,
            )

            return Apply(self, inputs, outputs)

        def perform(self, node, inputs, outputs):
            results = self.jitted_sol_op_jax(inputs)
            if self.num_outputs > 1:
                for i in range(self.num_outputs):
                    outputs[i][0] = np.array(results[i], self.output_types[i].dtype)
            else:
                outputs[0][0] = np.array(results, self.output_types[0].dtype)

        def perform_jax(self, *inputs):
            results = self.jitted_sol_op_jax(inputs)
            return results

        def grad(self, inputs, output_gradients):
            # If a output is not used, it is disconnected and doesn't have a gradient.
            # Set gradient here to zero for those outputs.
            for i in range(self.num_outputs):
                if isinstance(output_gradients[i].type, DisconnectedType):
                    if None not in self.output_types[i].shape:
                        output_gradients[i] = pt.zeros(
                            self.output_types[i].shape, self.output_types[i].dtype
                        )
                    else:
                        output_gradients[i] = pt.zeros((), self.output_types[i].dtype)
            result = self.vjp_sol_op(inputs, output_gradients)

            if self.num_inputs > 1:
                return result
            else:
                return (result,)  # Pytensor requires a tuple here

    # vector-jacobian product Op
    class VJPSolOp(Op):
        def __init__(
            self,
            input_treedef,
            input_types,
            jitted_vjp_sol_op_jax,
        ):
            self.input_treedef = input_treedef
            self.input_types = input_types
            self.jitted_vjp_sol_op_jax = jitted_vjp_sol_op_jax

        def make_node(self, y0, gz):
            y0 = [
                pt.as_tensor_variable(
                    _y,
                ).astype(self.input_types[i].dtype)
                for i, _y in enumerate(y0)
            ]
            gz_not_disconntected = [
                pt.as_tensor_variable(_gz)
                for _gz in gz
                if not isinstance(_gz.type, DisconnectedType)
            ]
            outputs = [in_type() for in_type in self.input_types]
            self.num_outputs = len(outputs)
            return Apply(self, y0 + gz_not_disconntected, outputs)

        def perform(self, node, inputs, outputs):
            results = self.jitted_vjp_sol_op_jax(tuple(inputs))
            if len(self.input_types) > 1:
                for i, result in enumerate(results):
                    outputs[i][0] = np.array(result, self.input_types[i].dtype)
            else:
                outputs[0][0] = np.array(results, self.input_types[0].dtype)

        def perform_jax(self, *inputs):
            results = self.jitted_vjp_sol_op_jax(tuple(inputs))
            if self.num_outputs == 1:
                if isinstance(results, Sequence):
                    return results[0]
                else:
                    return results
            else:
                return tuple(results)

    SolOp.__name__ = name
    SolOp.__qualname__ = ".".join(SolOp.__qualname__.split(".")[:-1] + [name])

    VJPSolOp.__name__ = "VJP_" + name
    VJPSolOp.__qualname__ = ".".join(
        VJPSolOp.__qualname__.split(".")[:-1] + ["VJP_" + name]
    )

    return SolOp, VJPSolOp



"""Tools to create compartmental models and ODE systems."""

import logging
from collections.abc import Sequence

import graphviz
import jax
import jax.numpy as jnp
from jaxtyping import Array, ArrayLike, PyTree

from .tree_tools import nested_indexing, walk_tree

# import jax.numpy as jnp

logger = logging.getLogger(__name__)


class CompModel:
    """Class to help building a compartmental model.

    The model is built by adding flows between compartments. The model is then compiled
    into a function that can be used in an ODE.

    Parameters
    ----------
    y_dict: PyTree[ArrayLike]
        Dictionary/nested python structure of compartments. Keys are the names of the
        compartments and values are ``jax.numpy.arrays`` that represent their value.

    Attributes
    ----------
    y: PyTree[jax.Array]
        Dictionary or pytree of compartments. Keys/indices are the names of the
        compartments and values are :class:`jax.Array` that represent their value.
    dy: PyTree[jax.Array]
        Dictionary or pytree of compartments. Keys/indices are the names of the
        compartments and values are :class:`jax.Array` that represent their
        derivative.
    edges: PyTree[list[tuple[indices, str]]]
        Edges between compartments, saved as the same pytree as the compartments,
        but as leafs, as a list of tuples. The first element of the tuple is the
        index of the compartment the flow is going to, the second element is the
        label of the flow.
    """

    def __init__(self, y_dict: PyTree[ArrayLike] = None):
        if y_dict is None:
            y_dict = {}
        self.y = y_dict

    @property
    def y(self) -> PyTree[Array]:
        """Get the compartments of the compartmental model."""
        return self._y

    @property
    def edges(self) -> PyTree[list[tuple[str | int | Sequence[str | int], str]]]:
        """Get the flow labels of the compartmental model."""
        return self._edges

    @y.setter
    def y(self, y_dict: PyTree[ArrayLike]) -> None:
        """Set the compartments of the model.

        Also resets the derivatives of the compartments to zero.
        """
        self._y = jax.tree_util.tree_map(lambda x: jnp.array(x), y_dict)
        self._dy = jax.tree_util.tree_map(lambda x: jnp.zeros_like(x), self.y)
        self._edges = jax.tree_util.tree_map(lambda _: [], y_dict)

    @property
    def dy(self):
        """Returns the derivative of the compartments."""
        return self._dy

    def flow(
        self,
        start_comp: str | int | Sequence[str | int],
        end_comp: str | int | Sequence[str | int],
        rate: ArrayLike,
        label: str | None = None,
        end_comp_is_erlang: bool = False,
    ) -> None:
        """
        Add a flow from ``start_comp`` to ``end_comp`` with rate flow.

        Parameters
        ----------
        start_comp:
            Key/index list of the start compartment.
        end_comp:
            Key/index list of the end compartment
        rate:
            rate of the flow to add between compartments, is multiplied by `start_comp`,
            so it should be broadcastable with it.
        label:
            label of the edge between the compartments that will be used when displaying
            a graph of the compartmental model.
        end_comp_is_erlang:
            If True, end_comp points to a compartment with an Erlang distributed
            dwelling time, i.e., as the last dimension of the compartment is used
            for the Erlang distribution modeling. The flow is then only added
            to the first element of the last dimension, i.e. to
            ``self.y[end_comp][...,0]``.

        Returns
        -------
        None
        """
        self.add_deriv(start_comp, -rate * nested_indexing(self.y, start_comp))
        self.add_deriv(
            end_comp, rate * nested_indexing(self.y, start_comp), end_comp_is_erlang
        )

        edges_list = nested_indexing(self.edges, start_comp)
        edges_list.append((end_comp, label))

    def add_deriv(
        self,
        y_key: str | int | Sequence[str | int],
        additive_dy: ArrayLike,
        end_comp_is_erlang: bool = False,
    ) -> None:
        """Add a derivative to a compartment.

        Add a derivative to a compartment. This is useful if the derivative is not
        directly modelled by a flow between compartments.

        Parameters
        ----------
        y_key:
            Key/index list of the compartment
        additive_dy:
            Derivative to add to the compartment
        end_comp_is_erlang:
            If True, ``y_key`` points to a compartment with an Erlang distributed
            dwelling time, i.e., as the last dimension of the compartment is used
            for the Erlang distribution modeling. The derivative is then only added
            to the first element of the last dimension, i.e. to
            ``self.y[end_comp][...,0]``

        """
        nested_indexing(
            tree=self.dy,
            indices=y_key,
            add=additive_dy,
            at=(Ellipsis, 0) if end_comp_is_erlang else None,
        )

    def erlang_flow(
        self,
        start_comp: str | int | Sequence[str | int],
        end_comp: str | int | Sequence[str | int],
        rate: ArrayLike,
        label: str | None = None,
        end_comp_is_erlang: bool = False,
    ) -> None:
        """Add a flow with erlang kernel from ``start_comp`` to ``end_comp``.

        Uses the function :func:`erlang_kernel` to model the flow.

        Parameters
        ----------
        start_comp:
            start compartments of the flow. The length of last dimension the list is
            the shape of the Erlang kernel.
        end_comp:
            end compartment of the flow
        rate:
            rate of the flow, equal to the inverse of the mean time spent in the Erlang
            kernel.
        label:
            label of the edge between the compartments that will be used when displaying
            a graph of the compartmental model.
        end_comp_is_erlang:
            If True, ``end_comp`` points to a compartment with an Erlang distributed
            dwelling time, i.e., as the last dimension of the compartment is used
            for the Erlang distribution modeling. The flow is then only added to the
            first element of the last dimension, i.e. to ``self.y[end_comp][...,0]``.
        """
        d_comp, outflow = erlang_kernel(
            comp=nested_indexing(self.y, start_comp), rate=rate
        )
        self.add_deriv(start_comp, d_comp)
        self.add_deriv(end_comp, outflow, end_comp_is_erlang)

        edges_list = nested_indexing(self.edges, start_comp)
        edges_list.append((end_comp, label))

    def delayed_copy(
        self,
        comp_to_copy: str | int | Sequence[str | int],
        delayed_comp: str | int | Sequence[str | int],
        tau_delay: ArrayLike,
    ) -> None:
        """Add a delayed copy of a compartment.

        Uses the function :func:`delayed_copy_kernel` to model the delayed copy.

        Parameters
        ----------
        comp_to_copy:
            Key/index list of the compartment to copy
        delayed_comp:
            Key/index list of the delayed compartment
        tau_delay:
            The mean delay of the copy, has to be broadcastable with the
            compartment``comp_to_copyp
        """
        d_delayed_comp = delayed_copy_kernel(
            initial_comp=nested_indexing(self.y, comp_to_copy),
            delayed_comp=nested_indexing(self.y, delayed_comp),
            tau_delay=tau_delay,
        )
        self.add_deriv(delayed_comp, d_delayed_comp)

    def view_graph(self, on_display: bool = True) -> None:
        """Display a graph of the compartmental model.

        Requires Graphviz (a non-python software) to be installed
        (https://www.graphviz.org/). It is also available in the conda-forge channel.
        See https://github.com/xflr6/graphviz?tab=readme-ov-file#installation.

        Parameters
        ----------
        on_display :
            If True, the graph is displayed in the notebook, otherwise it is saved as a
            pdf in the current folder and opened with the default pdf viewer.
        """
        graph = graphviz.Digraph("CompModel")
        graph.attr(rankdir="LR")

        for indices, _ in walk_tree(self.y):
            name_node = str(indices)[1:-1]
            graph.node(name_node)
        for indices, _ in walk_tree(self.y):
            name_startnode = str(indices)[1:-1]
            for edge in nested_indexing(self.edges, indices):
                endnode = edge[0] if isinstance(edge[0], list | tuple) else [edge[0]]
                name_endnode = str(endnode)[1:-1]
                graph.edge(name_startnode, name_endnode, label=edge[1])

        if on_display:
            try:
                from IPython.display import display

                display(graph)
            except Exception:
                graph.view()
        else:
            graph.view()


def delayed_copy_kernel(
    initial_comp: ArrayLike, delayed_comp: ArrayLike, tau_delay: ArrayLike
) -> Array:
    """Return the derivative to model the delayed copy of a compartment.

    The delay has the form of an Erlang kernel with shape parameter
    delayed_comp.shape[-1].

    Parameters
    ----------
    initial_comp :
        The compartment of shape (...) that is copied
    delayed_comp :
        Compartment of shape (..., n) that is a delayed copies of initial_var, the last
        element of the
        last dimension is the compartment which has the same total content as
        initial_comp over time, but is delayed by tau_delay.
    tau_delay :
        The mean delay of the copy, has to be broadcastable with ``initial_comp``.

    Returns
    -------
    d_delayed_vars :
        The derivatives, of shape (..., n) of the delayed compartments

    """
    length = jnp.shape(delayed_comp)[-1]
    inflow = initial_comp / tau_delay * length
    d_delayed_vars, outflow = erlang_kernel(inflow, delayed_comp[:], 1 / tau_delay)
    return d_delayed_vars


def erlang_kernel(
    comp: ArrayLike, rate: ArrayLike, inflow: ArrayLike = 0
) -> tuple[Array, ArrayLike]:
    r"""Model the compartments delayed by an Erlang kernel.

    Utility function to model an Erlang kernel for a compartmental model. The shape is
    determined by the length of the last dimension of comp. For example, if comp C
    is an array of shape (...,3), the function implements the following system of ODEs:

    .. math::
       \begin{align*}
       \mathrm{rate\_indiv} &= 3 \cdot \mathrm{rate},&\\
        \frac{\mathrm dC^{(1)}(t)}{\mathrm dt} &= \mathrm{inflow} &-\mathrm{
        rate\_indiv}
        \cdot
            C^{(1)},\\
        \frac{\mathrm dC^{(2)}(t)}{\mathrm dt} &= \mathrm{rate\_indiv}  \cdot
            C^{(1)} &-\mathrm{rate\_indiv}  \cdot C^{(2)},\\
        \frac{\mathrm dC^{(3)}(t)}{\mathrm dt} &= \mathrm{rate\_indiv}  \cdot
            C^{(2)} &-\mathrm{rate\_indiv}  \cdot C^{(3)},\\
        \mathrm{outflow} &= \mathrm{rate\_indiv}  \cdot C^{(3)}
       \end{align*}
    ..

    Parameters
    ----------
    comp:
        The compartment of shape (..., n) on which the Erlang kernel is applied. The
        last dimension n is the length of the kernel.
    rate:
        The rate of the kernel, 1/rate is the mean time spent in total in all
        compartments. Has to be broadcastable with ``comp``.

    Returns
    -------
    d_comp:
        The derivatives of shape (..., n) of the compartments
    outflow:
        The outflow of shape (...) from the last compartment
    """
    if not isinstance(comp, ArrayLike):
        raise AttributeError(
            f"comp has to be an array-like object, not of type {type(comp)}"
        )
    length = jnp.shape(comp)[-1]
    d_comp = jnp.zeros_like(comp)
    rate_indiv = rate * length
    d_comp = d_comp.at[..., 0].add(inflow - rate_indiv * comp[..., 0])
    d_comp = d_comp.at[..., 1:].add(
        -rate_indiv * comp[..., 1:] + rate_indiv * comp[..., :-1]
    )

    outflow = rate_indiv * comp[..., -1]

    return d_comp, outflow



"""Tests for the diffrax wrapper."""

import diffrax
import jax
import jax.numpy as jnp
import numpy as np
import pymc as pm
import pytest

import icomo
from icomo import interpolate_func, jax2pytensor


@pytest.fixture
def ode_models():
    with pm.Model() as model1:
        prey_0, predator_0 = pm.Normal("y0", size=2)
        args = pm.Normal("args", size=4)

        def vector_field(t, y, args):
            prey, predator = y
            α, β, γ, δ = args
            d_prey = α * prey - β * prey * predator
            d_predator = -γ * predator + δ * prey * predator
            d_y = d_prey, d_predator
            return d_y

        term = diffrax.ODETerm(vector_field)
        solver = diffrax.Tsit5()
        t0 = 0
        t1 = 14
        dt0 = 0.1
        y0 = (prey_0, predator_0)
        # args = (0.1, 0.02, 0.4, 0.02)
        saveat = diffrax.SaveAt(ts=jnp.linspace(t0, t1, 100))

        diffeqsolve = jax2pytensor(diffrax.diffeqsolve)

        sol = diffeqsolve(term, solver, t0, t1, dt0, y0=y0, args=args, saveat=saveat)
        pm.Normal("obs", sol.ys, observed=3 * np.ones((2, 100)))

    with pm.Model() as model2:
        y0 = pm.Normal("y0", size=2)
        args = pm.Normal("args", size=4)

        def vector_field(t, y, args):
            prey, predator = y
            α, β, γ, δ = args
            d_prey = α * prey - β * prey * predator
            d_predator = -γ * predator + δ * prey * predator
            d_y = jnp.array([d_prey, d_predator])
            return d_y

        term = diffrax.ODETerm(vector_field)
        solver = diffrax.Tsit5()
        t0 = 0
        t1 = 14
        dt0 = 0.1
        # args = (0.1, 0.02, 0.4, 0.02)
        # saveat = diffrax.SaveAt(ts=jnp.linspace(t0, t1, 100))
        diffeqsolve = jax2pytensor(diffrax.diffeqsolve)

        sol = diffeqsolve(term, solver, t0, t1, dt0, y0=y0, args=args, saveat=saveat)
        pm.Normal("obs", sol.ys, observed=np.array([[3, 3]]))

    def Erlang_SEIR_v2(t, y, args):
        beta_t_func = args["beta_t_func"]

        comp_model = icomo.CompModel(y)

        comp_model.flow(
            start_comp="S",
            end_comp="Es",
            rate=y["I"] / args["N"] * beta_t_func(t),
            label="beta(t) * I/N",  # label of the graph edge
            end_comp_is_erlang=True,
        )  # One has to specify that "Es" refers to
        # a list of compartments
        comp_model.erlang_flow(
            "Es", "I", args["rate_latent"], label="rate_latent (erlang)"
        )
        comp_model.flow("I", "R", args["rate_infectious"], label="rate_infectious")
        return comp_model.dy

    with pm.Model() as model3:
        len_sim = 30
        num_points = len_sim // 10

        t_out = np.linspace(0, len_sim, num_points)

        N = 1e5
        R0 = 1.5
        duration_latent = 3
        duration_infectious = 7
        beta0 = R0 / duration_infectious

        beta_t = pm.Normal("beta_t", mu=beta0, sigma=0.1, shape=len(t_out))
        beta_t_func = icomo.jax2pytensor(interpolate_func)(ts_in=t_out, values=beta_t)
        args = {
            "beta_t_func": beta_t_func,
            "N": N,
            "rate_latent": 1 / duration_latent,
            "rate_infectious": 1 / duration_infectious,
        }

        y0 = {
            "Es": np.array([100, 100, 100]),
            "I": pm.Normal("I0", mu=10, sigma=1),
            "R": 0,
        }
        # y0["S"] = N - jax.tree_util.tree_reduce(lambda x, y: x + y, y0)
        y0["S"] = N - y0["R"] - np.sum(y0["Es"])

        solution = icomo.jax2pytensor(icomo.diffeqsolve)(
            diffrax.ODETerm(Erlang_SEIR_v2),
            saveat=diffrax.SaveAt(ts=t_out),
            t0=min(t_out),
            t1=max(t_out),
            dt0=np.diff(t_out)[0],
            y0=y0,
            args=args,
            solver=diffrax.Tsit5(),
        )
        pm.Normal("obs", solution.ys["I"], observed=3 * np.ones(num_points))
        beta_t_interp = beta_t_func(t_out)
        pm.Deterministic("beta_t_interp", beta_t_interp)

    return (
        model1,
        model2,
        model3,
    )


def test_jax_compilation(ode_models):
    for i, model in enumerate(ode_models):
        print(f"ODE model {i + 1}")

        ip = model.initial_point()
        # Setting the mode to fast_compile shouldn't make a difference in the test
        # coverage
        logp_fn = model.compile_fn(model.logp(sum=False), mode="FAST_COMPILE")
        logp_fn(ip)
        dlogp_fn = model.compile_fn(model.dlogp(), mode="FAST_COMPILE")
        dlogp_fn(ip)

        ip = model.initial_point()
        logp_fn = model.compile_fn(model.logp(sum=False), mode="JAX")
        logp_fn(ip)
        dlogp_fn = model.compile_fn(model.dlogp(), mode="JAX")
        dlogp_fn(ip)


def test_jax_without_jit(ode_models):
    with jax.disable_jit():
        for i, model in enumerate(ode_models):
            print(f"Test model {i + 1}")

            ip = model.initial_point()
            logp_fn = model.compile_fn(model.logp(sum=False), mode="JAX")
            logp_fn(ip)
            dlogp_fn = model.compile_fn(model.dlogp(), mode="JAX")
            dlogp_fn(ip)



"""Tests for jax2pytensor package."""

import jax
import jax.numpy as jnp
import pymc as pm
import pytensor
import pytensor.tensor as pt
import pytest

from icomo import jax2pytensor


@pytest.fixture
def test_models():
    # 2 parameters input, tuple output
    with pm.Model() as model1:
        x, y = pm.Normal("input", size=2)

        @jax2pytensor
        def f(x, y):
            return jax.nn.sigmoid(x + y), y * 2

        out, _ = f(x, y)
        pm.Normal("obs", out, observed=3)

    # 2 parameters input, single output
    with pm.Model() as model2:
        x, y = pm.Normal("input", size=2)

        def f2(x, y):
            return jax.nn.sigmoid(x + y)

        f2_op = jax2pytensor(f2)
        out = f2_op(x, y)
        pm.Normal("obs", out, observed=3)

    # 2 parameters input, list output
    with pm.Model() as model3:
        x, y = pm.Normal("input", size=2)

        def f(x, y):
            return [jax.nn.sigmoid(x + y), y * 2]

        f_op = jax2pytensor(f)
        out, _ = f_op(x, y)
        pm.Normal("obs", out, observed=3)

    # single 1d input, tuple output
    with pm.Model() as model4:
        x, y = pm.Normal("input", size=2)

        def f4(x):
            return jax.nn.sigmoid(x), x * 2

        f4_op = jax2pytensor(f4)
        out, _ = f4_op(x)
        pm.Normal("obs", out, observed=3)

    # single 0d input, tuple output
    with pm.Model() as model5:
        x = pm.Normal("input", size=())

        def f5(x):
            return jax.nn.sigmoid(x), x

        f5_op = jax2pytensor(f5)
        out, _ = f5_op(x)
        pm.Normal("obs", out, observed=3)

    # single input, list output
    with pm.Model() as model6:
        x, y = pm.Normal("input", size=2)

        def f(x):
            return [jax.nn.sigmoid(x), 2 * x]

        f_op = jax2pytensor(f)
        out, _ = f_op(x)
        pm.Normal("obs", out, observed=3)

    # 2 parameters input with pytree, tuple output
    with pm.Model() as model7:
        x, y = pm.Normal("input", size=2)
        y_tmp = {"y": y, "y2": [y**2]}

        def f(x, y):
            return jax.nn.sigmoid(x), 2 * x + y["y"] + y["y2"][0]

        f_op = jax2pytensor(f)
        out, _ = f_op(x, y_tmp)
        pm.Normal("obs", out, observed=3)

    # 2 parameters input with pytree, pytree output
    with pm.Model() as model8:
        x = pm.Normal("input", size=3)
        y = pm.Normal("input2", size=(1,))
        y_tmp = {"a": y, "b": [y**2]}

        def f(x, y):
            jax.debug.print("x: {}", x)
            jax.debug.print("y: {}", y)
            return x, jax.tree_util.tree_map(
                lambda x: jnp.exp(x), y
            )  # {"a": y["y4"], "b": y["y3"]}  # , jax.tree_map(jnp.exp, y)

        f_op = jax2pytensor(f)
        out_x, out_y = f_op(x, y_tmp)
        # for_model = out_y["y3"]
        pm.Normal("obs", out_x, observed=(3, 2, 3))
        # pm.Normal("obs2", out_y["y3"], observed=(3,))

    # 2 parameters input with pytree, pytree output and non-graph argument
    with pm.Model() as model9:
        x = pm.Normal("input", size=3)
        y = pm.Normal("input2", size=1)
        y_tmp = {"a": y, "b": [y**2]}

        def f(x, y, non_model_arg):
            print(non_model_arg)
            return x, jax.tree_util.tree_map(jax.nn.sigmoid, y)

        f_op = jax2pytensor(
            f,
        )
        out_x, out_y = f_op(x, y_tmp, "Hello World!")

        pm.Normal("obs", out_y["b"][0], observed=(3,))

    # Use "None" in shape specification and have a non-used output of higher rank
    with pm.Model() as model10:
        x = pm.Normal("input", size=3)
        y = pm.Normal("input2", size=3)

        def f(x, y):
            return x[:, None] @ y[None], x

        f_op = jax2pytensor(f)
        out_x, out_y = f_op(x, y)

        pm.Normal("obs", out_y[0], observed=(3,))

    with pm.Model() as model11:
        x = pm.Normal("input", size=3)
        y = pm.Normal("input2", size=3)

        # Now x has an unknown shape
        x, _ = pytensor.map(pt.exp, [x])
        x.type.shape = (3,)

        def f11(x, y):
            return x[:, None] @ y[None]

        f_op = jax2pytensor(f11)
        out_x = f_op(x, y)
        # out_x = x[:, None] @ y[None]

        pm.Normal("obs", out_x[0, 1], observed=2)

    # Test broadcasting with unknown shape
    with pm.Model() as model12:
        x = pm.Normal("input", size=3)
        y = pm.Normal("input2", size=3)

        shape_before = x.type.shape

        # Now x has an unknown shape
        x = pt.cumsum(x)

        x.type.shape = shape_before

        def f12(x, y):
            return x * jnp.ones(3)

        f_op = jax2pytensor(f12)
        out_x = f_op(x, y)
        # out_x = x[:, None] @ y[None]

        pm.Normal("obs", out_x[0], observed=2)

    with pm.Model() as model13:
        x = pm.Normal("input", size=3)
        y = pm.Normal("input2", size=3)

        shape_before = x.type.shape

        # Now x has an unknown shape
        x = pt.cumsum(x)

        def f12(x, y):
            return x * jnp.ones(3)

        f_op = jax2pytensor(f12)
        out_x = f_op(x, y)
        # out_x = x[:, None] @ y[None]

        pm.Normal("obs", out_x[0], observed=2)

    # Test non-array return values
    with pm.Model() as model14:
        x = pm.Normal("input", size=3)
        y = pm.Normal("input2", size=3)

        # Now x has an unknown shape
        x = pt.cumsum(x)

        def f14(x, y, message):
            return x * jnp.ones(3), "Success: " + message

        f_op = jax2pytensor(f14)
        out_x, message = f_op(x, y, "Hi")
        assert message == "Success: Hi"

        pm.Normal("obs", out_x[0], observed=2)

    # Test nested functions
    with pm.Model() as model15:
        x = pm.Normal("input", size=3)
        y = pm.Normal("input2", size=3)

        @jax2pytensor
        def f_internal(y):
            def f_ret(t):
                return y + t

            def f_ret2(t):
                return f_ret(t) + t**2

            return f_ret, y**2, f_ret2

        f, y_pow, f2 = f_internal(y)

        @jax2pytensor
        def f15(x, dict_other):
            f, y_pow = dict_other["func"], dict_other["y"]
            return x * jnp.ones(3), f(x) * y_pow

        out_x = f15(x, {"func": f, "y": y_pow})

        x_res = out_x[1] + f(y * 3) + f2(x)

        pm.Normal("obs", x_res, observed=[2, 2, 2])

    return (
        model1,
        model2,
        model3,
        model4,
        model5,
        model6,
        model7,
        model8,
        model9,
        model10,
        model11,
        model12,
        model13,
        model14,
        model15,
    )


def test_jax_compilation(test_models):
    for i, model in enumerate(test_models):
        print(f"Test model {i + 1}")

        ip = model.initial_point()
        # Setting the mode to fast_compile shouldn't make a difference in the test
        # coverage
        logp_fn = model.compile_fn(model.logp(sum=False), mode="FAST_COMPILE")
        logp_fn(ip)
        dlogp_fn = model.compile_fn(model.dlogp(), mode="FAST_COMPILE")
        dlogp_fn(ip)

        ip = model.initial_point()
        logp_fn = model.compile_fn(model.logp(sum=False), mode="JAX")
        logp_fn(ip)
        dlogp_fn = model.compile_fn(model.dlogp(), mode="JAX")
        dlogp_fn(ip)


def test_jax_without_jit(test_models):
    with jax.disable_jit():
        for i, model in enumerate(test_models):
            print(f"Test model {i + 1}")

            ip = model.initial_point()
            logp_fn = model.compile_fn(model.logp(sum=False), mode="JAX")
            logp_fn(ip)
            dlogp_fn = model.compile_fn(model.dlogp(), mode="JAX")
            dlogp_fn(ip)
